import typing as t
from abc import ABC, abstractmethod

from .base import ModelEvent


class EventHandler(ABC):
    @abstractmethod
    def __call__(self, events: t.Sequence[ModelEvent[t.Any]]) -> None:
        raise NotImplementedError



import typing as t
from abc import ABC, abstractmethod

from ...base import ModelEvent


class EventHandler(ABC):
    @abstractmethod
    async def __call__(self, events: t.Sequence[ModelEvent[t.Any]]) -> None:
        raise NotImplementedError



from .base import (
    AuditedEntity,
    Entity,
    SoftDeletableEntity,
    Version,
    VersionedEntity,
)

__all__ = (
    "AuditedEntity",
    "Entity",
    "SoftDeletableEntity",
    "Version",
    "VersionedEntity",
)



import datetime


def offset_naive_utcnow() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)


def nano_timestamp_utc() -> int:
    return int(offset_naive_utcnow().timestamp() * 1e9)



import datetime
import typing as t
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field, replace
from enum import Enum, auto, unique

from ..clock import offset_naive_utcnow
from ..entity import (
    AuditedEntity,
    Entity,
    SoftDeletableEntity,
    VersionedEntity,
)
from ..types import MISSING
from .event import ModelCreatedEvent, ModelDeletedEvent, ModelEvent

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound=Entity[t.Any])
SELF = t.TypeVar("SELF", bound="Model[t.Any]")


@dataclass(frozen=True)
class Model(
    AuditedEntity[ENTITY_ID],
    SoftDeletableEntity[ENTITY_ID],
    VersionedEntity[ENTITY_ID],
    metaclass=ABCMeta,
):
    id: ENTITY_ID = t.cast(ENTITY_ID, MISSING)

    _events: t.Sequence[ModelEvent[ENTITY_ID]] = field(
        default_factory=list, repr=False, compare=False
    )
    _is_new: bool = field(default=False, repr=False, compare=False)

    @abstractmethod
    def _generate_id(self) -> ENTITY_ID:
        raise NotImplementedError

    @abstractmethod
    def _created_event(
        self, date: datetime.datetime
    ) -> ModelCreatedEvent[ENTITY_ID]:
        raise NotImplementedError

    @abstractmethod
    def _deleted_event(
        self, date: datetime.datetime
    ) -> ModelDeletedEvent[ENTITY_ID]:
        raise NotImplementedError

    def __post_init__(self) -> None:
        super().__post_init__()

        if self.id is MISSING:
            object.__setattr__(self, "id", self._generate_id())
            object.__setattr__(self, "_is_new", True)
            object.__setattr__(
                self, "_events", [self._created_event(self.created_date)]
            )

    def update(
        self: SELF,
        event: t.Optional[ModelEvent[ENTITY_ID]] = None,
        **kwargs: t.Any,
    ) -> SELF:
        return replace(
            self,
            _events=[*self._events, event] if event else self._events,
            **kwargs,
        )

    def delete(self: SELF) -> SELF:
        now = offset_naive_utcnow()
        return self.update(self._deleted_event(now), deleted_date=now)

    @property
    def is_new(self) -> bool:
        return self._is_new

    @property
    def is_deleted(self) -> bool:
        return self.deleted_date is not None

    def events(self) -> t.Sequence[ModelEvent[ENTITY_ID]]:
        return deepcopy(self._events)


@unique
class ChangeType(Enum):
    ADD = auto()
    UPDATE = auto()
    DELETE = auto()


@dataclass(frozen=True)
class Change:
    type: ChangeType
    entity: Entity[t.Any]


@dataclass(frozen=True)
class Batch:
    is_shut: bool = field(default=False, init=False)

    _changes: t.Dict[t.Any, Change] = field(
        default_factory=dict, init=False, repr=False
    )

    def events(self) -> t.Sequence[ModelEvent[t.Any]]:
        return [
            event
            for change in self._changes.values()
            for event in (
                change.entity.events()
                if isinstance(change.entity, Model)
                else []
            )
        ]

    def changes(self) -> t.Dict[ENTITY_ID, Change]:
        return deepcopy(self._changes)

    def add(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        if isinstance(entity, Model) and not entity.is_new:
            raise RuntimeError("Can't add an existing entity")

        self._add_change(ChangeType.ADD, entity)
        return entity

    def update(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        if isinstance(entity, Model) and entity.is_new:
            raise RuntimeError("Can't update a new entity")

        self._add_change(ChangeType.UPDATE, entity)
        return entity

    def delete(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        if isinstance(entity, Model) and entity.is_new:
            raise RuntimeError("Can't delete a new entity")

        self._add_change(ChangeType.DELETE, entity)
        return entity

    def shut(self) -> None:
        object.__setattr__(self, "is_shut", True)

    def _add_change(
        self, change_type: ChangeType, entity: ENTITY_TYPE
    ) -> None:
        if self.is_shut:
            raise RuntimeError("Batch can't be changed, it's already shut")

        if existing := self._changes.get(entity.id):
            raise RuntimeError(
                f"{entity.__class__.__name__}[{entity.id}] has already been added to batch with {existing.type.name} operation"
            )

        self._changes[entity.id] = Change(type=change_type, entity=entity)



import typing as t
from abc import ABC
from itertools import groupby

from ....domain import Batch, ChangeType, Model
from ....domain.event import ModelEvent
from ....repository.aio import BaseRepositoryFactory
from ....repository.aio.base import ENTITY_TYPE


class BaseDomainRepository(ABC):
    def __init__(
        self,
        *,
        repositories: BaseRepositoryFactory,
        events_handler: t.Callable[
            [t.Sequence[ModelEvent[t.Any]]], t.Awaitable[None]
        ],
    ) -> None:
        self._repositories = repositories
        self._events_handler = events_handler

    async def process_batch(self, batch: Batch) -> None:
        batch.shut()

        for (change_type, _), changes in groupby(
            batch.changes().values(), lambda c: (c.type, type(c.entity))
        ):
            entities = [change.entity for change in changes]

            if change_type == ChangeType.ADD:
                await self._add_all(entities)
            if change_type == ChangeType.UPDATE:
                await self._update_all(entities)
            if change_type == ChangeType.DELETE:
                await self._delete_all(entities)

        await self._events_handler(batch.events())

    async def _add(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        return t.cast(
            ENTITY_TYPE,
            await self._repositories.repo_for(type(entity)).add(entity),
        )

    async def _add_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [await self._add(entity) for entity in entities]

    async def _update(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        return t.cast(
            ENTITY_TYPE,
            await self._repositories.repo_for(type(entity)).update(entity),
        )

    async def _update_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [await self._update(entity) for entity in entities]

    async def _delete(self, entity: ENTITY_TYPE) -> bool:
        deleted_entity = (
            (entity if entity.is_deleted else entity.delete())
            if isinstance(entity, Model)
            else entity
        )
        return await self._repositories.repo_for(type(deleted_entity)).delete(
            deleted_entity
        )

    async def _delete_all(self, entities: t.Sequence[ENTITY_TYPE]) -> bool:
        return all([await self._delete(entity) for entity in entities])



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ...datapoint import (
    BaseDataPointContainer,
    BaseDataPointsDict,
    BaseDataPointSpec,
    DataPointIsNotProducedError,
)
from ...datapoint.aio import BaseDataPointConsumer, BaseDataPointProducer
from ..exceptions import DataPointIsNotDeclaredError


class ConsumesDataPoints(ABC):
    @property
    @abstractmethod
    def _consumes(self) -> t.Set[BaseDataPointSpec[t.Any]]:
        raise NotImplementedError

    async def out_of(
        self, consumer: BaseDataPointConsumer[t.Any]
    ) -> BaseDataPointsDict:
        consumed = await consumer.get(*self._consumes)
        consumed_specs = set(consumed.keys())

        if not self._consumes.issubset(consumed_specs):
            missing_specs = self._consumes - consumed_specs
            raise DataPointIsNotProducedError(missing_specs)

        return t.cast(BaseDataPointsDict, consumed)


class ProducesDataPoints(ABC):
    @dataclass(frozen=True)
    class ProducerProxy:
        _producer: BaseDataPointProducer[t.Any]
        _required_specs: t.Set[BaseDataPointSpec[t.Any]]

        async def add(
            self, *datapoints: BaseDataPointContainer[t.Any]
        ) -> None:
            actual_specs = {datapoint.spec for datapoint in datapoints}
            missing_specs = self._required_specs - actual_specs
            extra_specs = actual_specs - self._required_specs

            if len(missing_specs) > 0:
                raise DataPointIsNotProducedError(missing_specs)

            if len(extra_specs) > 0:
                raise DataPointIsNotDeclaredError(extra_specs)

            await self._producer.add(*datapoints)

    @property
    @abstractmethod
    def _produces(self) -> t.Set[BaseDataPointSpec[t.Any]]:
        raise NotImplementedError

    def to(self, producer: BaseDataPointProducer[t.Any]) -> ProducerProxy:
        return self.ProducerProxy(producer, self._produces)



import datetime
import typing as t
from dataclasses import dataclass, field
from uuid import UUID

from ...clock import nano_timestamp_utc
from ...entity import Entity

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound=Entity[t.Any])


@dataclass(frozen=True)
class ModelEvent(t.Generic[ENTITY_ID]):
    id: UUID
    model_id: ENTITY_ID
    occurred_at: int = field(
        default_factory=nano_timestamp_utc,
        init=False,
        repr=False,
        compare=False,
    )


@dataclass(frozen=True)
class ModelCreatedEvent(ModelEvent[ENTITY_ID]):
    created_date: datetime.datetime


@dataclass(frozen=True)
class ModelDeletedEvent(ModelEvent[ENTITY_ID]):
    deleted_date: datetime.datetime



import typing as t
from abc import ABC, abstractmethod
from dataclasses import asdict

from .....clock import offset_naive_utcnow
from ...tables import SoftDeletableEntityTable

try:
    from sqlalchemy import (
        ColumnElement,
        delete,
        exists,
        insert,
        select,
        update,
    )
except ImportError:  # pragma: no cover
    raise ImportError(
        "Seems that you are trying to import extra module that was not installed,"
        " please install pyuow[sqlalchemy]"
    )

from .....contrib.sqlalchemy.aio.work.impl import (
    SqlAlchemyReadOnlyTransactionManager,
    SqlAlchemyTransactionManager,
)
from .....contrib.sqlalchemy.tables import (
    AuditedEntityTable,
    EntityTable,
    VersionedEntityTable,
)
from .....entity import (
    AuditedEntity,
    Entity,
    SoftDeletableEntity,
    VersionedEntity,
)
from .....repository.aio import BaseEntityRepository, BaseRepositoryFactory

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound=Entity[t.Any])
ENTITY_TABLE = t.TypeVar("ENTITY_TABLE", bound=EntityTable)


class BaseSqlAlchemyEntityRepository(
    t.Generic[ENTITY_ID, ENTITY_TYPE, ENTITY_TABLE],
    BaseEntityRepository[ENTITY_ID, ENTITY_TYPE],
    ABC,
):
    def __init__(
        self,
        table: t.Type[ENTITY_TABLE],
        transaction_manager: SqlAlchemyTransactionManager,
        readonly_transaction_manager: SqlAlchemyReadOnlyTransactionManager,
    ) -> None:
        self._table = table
        self._transaction_manager = transaction_manager
        self._readonly_transaction_manager = readonly_transaction_manager

    @staticmethod
    @abstractmethod
    def to_entity(record: ENTITY_TABLE) -> ENTITY_TYPE:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_record(entity: ENTITY_TYPE) -> ENTITY_TABLE:
        raise NotImplementedError

    async def find(self, entity_id: ENTITY_ID) -> t.Optional[ENTITY_TYPE]:
        statement = select(self._table).where(
            self._table.id == entity_id, *self._exclude_deleted()
        )

        async with self._readonly_transaction_manager.transaction() as trx:
            result = (await trx.it().execute(statement)).scalar_one_or_none()

        return self.to_entity(result) if result else None

    async def find_all(
        self, entity_ids: t.Iterable[ENTITY_ID]
    ) -> t.Iterable[ENTITY_TYPE]:
        statement = select(self._table).where(
            self._table.id.in_(entity_ids), *self._exclude_deleted()
        )

        async with self._readonly_transaction_manager.transaction() as trx:
            result = (await trx.it().execute(statement)).scalars().all()

        return [self.to_entity(record) for record in result]

    async def get(self, entity_id: ENTITY_ID) -> ENTITY_TYPE:
        statement = select(self._table).where(
            self._table.id == entity_id, *self._exclude_deleted()
        )

        async with self._readonly_transaction_manager.transaction() as trx:
            result = (await trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    async def exists(self, entity_id: ENTITY_ID) -> bool:
        statement = (
            exists(self._table)
            .where(self._table.id == entity_id, *self._exclude_deleted())
            .select()
        )

        async with self._readonly_transaction_manager.transaction() as trx:
            return (await trx.it().execute(statement)).scalar() or False

    async def add(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        statement = (
            insert(self._table)
            .values(**asdict(self.to_record(entity)))
            .returning(self._table)
        )

        async with self._transaction_manager.transaction() as trx:
            result = (await trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    async def add_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [await self.add(entity) for entity in entities]

    async def update(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        record = self.to_record(entity)
        conditions = self._exclude_deleted()

        if (
            isinstance(record, AuditedEntityTable)
            and issubclass(self._table, AuditedEntityTable)
            and isinstance(entity, AuditedEntity)
        ):
            record.updated_date = offset_naive_utcnow()
            conditions.append(self._table.created_date == entity.created_date)
            conditions.append(self._table.updated_date == entity.updated_date)

        if (
            isinstance(record, VersionedEntityTable)
            and issubclass(self._table, VersionedEntityTable)
            and isinstance(entity, VersionedEntity)
        ):
            record.version = entity.version.next()
            conditions.append(self._table.version == entity.version)

        statement = (
            update(self._table)
            .values(**asdict(record))
            .where(self._table.id == entity.id, *conditions)
            .returning(self._table)
        )

        async with self._transaction_manager.transaction() as trx:
            result = (await trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    async def update_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [await self.update(entity) for entity in entities]

    async def delete(self, entity: ENTITY_TYPE) -> bool:
        statement = (
            (
                update(self._table)
                .values(
                    deleted_date=(
                        entity.deleted_date
                        if entity.deleted_date
                        else offset_naive_utcnow()
                    )
                )
                .where(self._table.id == entity.id, *self._exclude_deleted())
                .returning(self._table.id)
            )
            if isinstance(entity, SoftDeletableEntity)
            else (
                delete(self._table)
                .where(self._table.id == entity.id, *self._exclude_deleted())
                .returning(self._table.id)
            )
        )

        async with self._transaction_manager.transaction() as trx:
            identifier = (await trx.it().execute(statement)).scalar_one()

        return t.cast(bool, identifier == entity.id)

    async def delete_all(self, entities: t.Sequence[ENTITY_TYPE]) -> bool:
        return all([await self.update(entity) for entity in entities])

    def _exclude_deleted(self) -> t.List[ColumnElement[bool]]:
        conditions: t.List[ColumnElement[bool]] = []

        if issubclass(self._table, SoftDeletableEntityTable):
            conditions.append(self._table.deleted_date == None)

        return conditions


class BaseSqlAlchemyRepositoryFactory(BaseRepositoryFactory, ABC):
    def __init__(
        self,
        transaction_manager: SqlAlchemyTransactionManager,
        readonly_transaction_manager: SqlAlchemyReadOnlyTransactionManager,
    ) -> None:
        self._transaction_manager = transaction_manager
        self._readonly_transaction_manager = readonly_transaction_manager



from .base import ModelCreatedEvent, ModelDeletedEvent, ModelEvent
from .dist import EventHandler

__all__ = (
    "ModelCreatedEvent",
    "ModelDeletedEvent",
    "ModelEvent",
    "EventHandler",
)



import typing as t
from dataclasses import dataclass, field

from ...context import BaseContext
from ...domain import Batch
from ..base import BaseParams

PARAMS = t.TypeVar("PARAMS", bound=BaseParams)


@dataclass(frozen=True)
class BaseDomainContext(BaseContext[PARAMS]):
    params: PARAMS
    batch: Batch = field(default_factory=Batch)



import typing as t

from .....context import BaseContext
from .....context.domain import BaseDomainContext
from .....domain import Batch
from .....result import Result
from .....unit.aio import BaseUnit
from .....work.aio import BaseUnitProxy
from .....work.aio.transactional import (
    BaseTransactionManager,
    TransactionalWorkManager,
)
from .. import BaseTransaction

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")
TRANSACTION = t.TypeVar("TRANSACTION", bound=BaseTransaction[t.Any])


class DomainUnit(BaseUnit[CONTEXT, OUT]):
    def __init__(
        self,
        *,
        unit: BaseUnit[CONTEXT, OUT],
        batch_handler: t.Callable[[Batch], t.Awaitable[None]],
    ) -> None:
        self._unit = unit
        self._batch_handler = batch_handler

    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        result = await self._unit(context)

        if isinstance(context, BaseDomainContext):
            await self._batch_handler(context.batch)

        return result


class DomainTransactionalWorkManager(TransactionalWorkManager):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
        batch_handler: t.Callable[[Batch], t.Awaitable[None]],
    ) -> None:
        super().__init__(transaction_manager=transaction_manager)
        self._batch_handler = batch_handler

    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        return super().by(
            unit=DomainUnit(
                unit=unit,
                batch_handler=self._batch_handler,
            )
        )



import typing as t

from ....context import BaseContext
from ....context.domain import BaseDomainContext
from ....domain import Batch
from ....result import Result
from ....unit import BaseUnit
from ....work import BaseUnitProxy
from ....work.transactional import (
    BaseTransactionManager,
    TransactionalWorkManager,
)
from .. import BaseTransaction

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")
TRANSACTION = t.TypeVar("TRANSACTION", bound=BaseTransaction[t.Any])


class DomainUnit(BaseUnit[CONTEXT, OUT]):
    def __init__(
        self,
        *,
        unit: BaseUnit[CONTEXT, OUT],
        batch_handler: t.Callable[[Batch], None],
    ) -> None:
        self._unit = unit
        self._batch_handler = batch_handler

    def __call__(self, context: CONTEXT) -> Result[OUT]:
        result = self._unit(context)

        if isinstance(context, BaseDomainContext):
            self._batch_handler(context.batch)

        return result


class DomainTransactionalWorkManager(TransactionalWorkManager):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
        batch_handler: t.Callable[[Batch], None],
    ) -> None:
        super().__init__(transaction_manager=transaction_manager)
        self._batch_handler = batch_handler

    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        return super().by(
            unit=DomainUnit(
                unit=unit,
                batch_handler=self._batch_handler,
            )
        )
